@@ -24,7 +24,7 @@ class Packet(object):
         '''Perform any parsing necessary to populate fields on this packet.
         This method is called as the last step in __init__(). The base class
         method is a no-op; subclasses should use this as required.'''
-        pass
+        return 0
 
     def __repr__(self):
         new = "old"
@@ -73,6 +73,36 @@ class AlgoLookup(object):
             return "Private/Experimental algorithm"
         return cls.hash_algorithms.get(alg, "Unknown")
 
+    sym_algorithms = {
+        # (Name, IV length)
+        0:  ("Plaintext or unencrypted", 0),
+        1:  ("IDEA", 8),
+        2:  ("Triple-DES", 8),
+        3:  ("CAST5", 8),
+        4:  ("Blowfish", 8),
+        5:  ("Reserved", 8),
+        6:  ("Reserved", 8),
+        7:  ("AES with 128-bit key", 16),
+        8:  ("AES with 192-bit key", 16),
+        9:  ("AES with 256-bit key", 16),
+        10: ("Twofish with 256-bit key", 16),
+        11: ("Camellia with 128-bit key", 16),
+        12: ("Camellia with 192-bit key", 16),
+        13: ("Camellia with 256-bit key", 16),
+    }
+
+    @classmethod
+    def _lookup_sym_algorithm(cls, alg):
+        return cls.sym_algorithms.get(alg, ("Unknown", 0))
+
+    @classmethod
+    def lookup_sym_algorithm(cls, alg):
+        return cls._lookup_sym_algorithm(alg)[0]
+
+    @classmethod
+    def lookup_sym_algorithm_iv(cls, alg):
+        return cls._lookup_sym_algorithm(alg)[1]
+
 
 class SignatureSubpacket(object):
     '''A signature subpacket containing a type, type name, some flags, and the
@@ -216,6 +246,8 @@ class SignaturePacket(Packet, AlgoLookup):
             raise PgpdumpException("Unsupported signature packet, version %d" %
                     self.sig_version)
 
+        return offset
+
     def parse_subpackets(self, outer_offset, outer_length, hashed=False):
         offset = outer_offset
         while offset < outer_offset + outer_length:
@@ -369,6 +401,8 @@ class PublicKeyPacket(Packet, AlgoLookup):
             raise PgpdumpException("Unsupported public key packet, version %d" %
                     self.pubkey_version)
 
+        return offset
+
     def parse_key_material(self, offset):
         if self.raw_pub_algorithm in (1, 2, 3):
             self.pub_algorithm_type = "rsa"
@@ -418,6 +452,161 @@ class PublicSubkeyPacket(PublicKeyPacket):
     pass
 
 
+class SecretKeyPacket(PublicKeyPacket):
+    def __init__(self, *args, **kwargs):
+        self.s2k_id = None
+        self.s2k_type = None
+        self.s2k_cipher = None
+        self.s2k_hash = None
+        self.s2k_iv = None
+        self.checksum = None
+        # RSA fields
+        self.exponent_d = None
+        self.prime_p = None
+        self.prime_q = None
+        self.multiplicative_inverse = None
+        # DSA and Elgamal
+        self.exponent_x = None
+        super(SecretKeyPacket, self).__init__(*args, **kwargs)
+
+    def parse(self):
+        # parse the public part
+        offset = PublicKeyPacket.parse(self)
+
+        # parse secret-key packet format from section 5.5.3
+        self.s2k_id = self.data[offset]
+        offset += 1
+
+        if self.s2k_id == 0:
+            # plaintext key data
+            offset = self.parse_private_key_material(offset)
+            self.checksum = get_int2(self.data, offset)
+            offset += 2
+        elif self.s2k_id in (254, 255):
+            # encrypted key data
+            cipher_id = self.data[offset]
+            offset += 1
+            self.s2k_cipher = self.lookup_sym_algorithm(cipher_id)
+
+            # s2k_length is the len of the entire S2K specifier, as per
+            # section 3.7.1 in RFC 4880
+            # we parse the info inside the specifier, but verify the # of
+            # octects we've parsed matches the expected length of the s2k
+            offset_before_s2k = offset
+
+            s2k_type_id = self.data[offset]
+            offset += 1
+            name, s2k_length = S2K_TYPES.get(s2k_type_id, ("Unknown", 0))
+            self.s2k_type = name
+
+            has_iv = True
+            if s2k_type_id == 0:
+                # simple string-to-key
+                hash_id = self.data[offset]
+                offset += 1
+                self.s2k_hash = self.lookup_hash_algorithm(hash_id)
+
+            elif s2k_type_id == 1:
+                # salted string-to-key
+                hash_id = self.data[offset]
+                offset += 1
+                self.s2k_hash = self.lookup_hash_algorithm(hash_id)
+                # ignore 8 bytes
+                offset += 8
+
+            elif s2k_type_id == 2:
+                # reserved
+                pass
+
+            elif s2k_type_id == 3:
+                # iterated and salted
+                hash_id = self.data[offset]
+                offset += 1
+                self.s2k_hash = self.lookup_hash_algorithm(hash_id)
+                # ignore 8 bytes
+                offset += 8
+                # ignore count
+                offset += 1
+                # TODO: parse and store count ?
+
+            elif 100 <= s2k_type_id <= 110:
+                # GnuPG string-to-key
+                # According to g10/parse-packet.c near line 1832, the 101 packet
+                # type is a special GnuPG extension.  This S2K extension is
+                # 6 bytes in total:
+                #
+                #   Octet 0:   101
+                #   Octet 1:   hash algorithm
+                #   Octet 2-4: "GNU"
+                #   Octet 5:   mode integer
+                hash_id = self.data[offset]
+                offset += 1
+                self.s2k_hash = self.lookup_hash_algorithm(hash_id)
+
+                gnu = self.data[offset:offset + 3]
+                offset += 3
+                if gnu != bytearray(b"GNU"):
+                    raise PgpdumpException(
+                            "S2K parsing error: expected 'GNU', got %s" % gnu)
+
+                mode = self.data[offset]
+                mode += 1000
+                offset += 1
+                if mode == 1001:
+                    has_iv = False
+                else:
+                    # TODO implement other modes?
+                    raise PgpdumpException(
+                            "Unsupported GnuPG S2K extension, encountered mode %d" % mode)
+            else:
+                raise PgpdumpException(
+                        "Unsupported public key algorithm %d" % s2k_type_id)
+
+            if s2k_length != (offset - offset_before_s2k):
+                raise PgpdumpException(
+                        "Error parsing string-to-key specifier, mismatched length")
+
+            if has_iv:
+                s2k_iv_len = self.lookup_sym_algorithm_iv(cipher_id)
+                self.s2k_iv = self.data[offset:offset + s2k_iv_len]
+                offset += s2k_iv_len
+
+            # TODO decrypt key data
+            # TODO parse checksum
+            return offset
+
+    def parse_private_key_material(self, offset):
+        if self.raw_pub_algorithm in (1, 2, 3):
+            self.pub_algorithm_type = "rsa"
+            # d, p, q, u
+            self.exponent_d, offset = get_mpi(self.data, offset)
+            self.prime_p, offset = get_mpi(self.data, offset)
+            self.prime_q, offset = get_mpi(self.data, offset)
+            self.multiplicative_inverse, offset = get_mpi(self.data, offset)
+        elif self.raw_pub_algorithm == 17:
+            self.pub_algorithm_type = "dsa"
+            # x
+            self.exponent_x, offset = get_mpi(self.data, offset)
+        elif self.raw_pub_algorithm in (16, 20):
+            self.pub_algorithm_type = "elg"
+            # x
+            self.exponent_x, offset = get_mpi(self.data, offset)
+        elif 100 <= self.raw_pub_algorithm <= 110:
+            # Private/Experimental algorithms, just move on
+            pass
+        else:
+            raise PgpdumpException("Unsupported public key algorithm %d" %
+                    self.raw_pub_algorithm)
+
+        return offset
+
+
+class SecretSubkeyPacket(SecretKeyPacket):
+    '''A Secret-Subkey packet (tag 7) has exactly the same format as a
+    Secret-Key packet, but denotes a subkey.'''
+    pass
+
+
 class UserIDPacket(Packet):
     '''A User ID packet consists of UTF-8 text that is intended to represent
     the name and email address of the key holder. By convention, it includes an
@@ -439,6 +628,8 @@ class UserIDPacket(Packet):
             if matches.group(2):
                 self.user_email = matches.group(2).strip()
 
+        return self.length
+
     def __repr__(self):
         return "<%s: %r (%r), length %d>" % (
                 self.__class__.__name__, self.user_name, self.user_email,
@@ -454,7 +645,7 @@ class UserAttributePacket(Packet):
 
     def parse(self):
         offset = sub_offset = sub_len = 0
-        while offset + sub_len < len(self.data):
+        while offset + sub_len < self.length:
             # each subpacket is [variable length] [subtype] [data]
             sub_offset, sub_len, sub_part = new_tag_length(self.data, offset)
             # sub_len includes the subtype single byte, knock that off
@@ -479,6 +670,8 @@ class UserAttributePacket(Packet):
                 else:
                     self.image_format = "unknown"
 
+        return self.length
+
 
 class TrustPacket(Packet):
     def __init__(self, *args, **kwargs):
@@ -490,18 +683,50 @@ class TrustPacket(Packet):
         integer values into some internal enumeration.'''
         if self.length == 2:
             self.trust = get_int2(self.data, 0)
+            return 2
+        return 0
+
+
+class PublicKeyEncryptedSessionKeyPacket(Packet, AlgoLookup):
+    def __init__(self, *args, **kwargs):
+        self.session_key_version = None
+        self.key_id = None
+        self.raw_pub_algorithm = None
+        self.pub_algorithm = None
+        super(PublicKeyEncryptedSessionKeyPacket, self).__init__(
+                *args, **kwargs)
+
+    def parse(self):
+        self.session_key_version = self.data[0]
+        if self.session_key_version == 3:
+            self.key_id = get_key_id(self.data, 1)
+            self.raw_pub_algorithm = self.data[9]
+            self.pub_algorithm = self.lookup_pub_algorithm(self.raw_pub_algorithm)
+        else:
+            raise PgpdumpException(
+                    "Unsupported encrypted session key packet, version %d" %
+                    self.session_key_version)
+
+        # this is hardcoded to work with the only known session key version
+        return 10
+
+    def __repr__(self):
+        return "<%s: 0x%s (%s), length %d>" % (
+                self.__class__.__name__, self.key_id, self.pub_algorithm,
+                self.length)
 
 
 TAG_TYPES = {
     # (Name, PacketType) tuples
     0:  ("Reserved", None),
-    1:  ("Public-Key Encrypted Session Key Packet", None),
+    1:  ("Public-Key Encrypted Session Key Packet",
+         PublicKeyEncryptedSessionKeyPacket),
     2:  ("Signature Packet", SignaturePacket),
     3:  ("Symmetric-Key Encrypted Session Key Packet", None),
     4:  ("One-Pass Signature Packet", None),
-    5:  ("Secret Key Packet", None),
+    5:  ("Secret Key Packet", SecretKeyPacket),
     6:  ("Public Key Packet", PublicKeyPacket),
-    7:  ("Secret Subkey Packet", None),
+    7:  ("Secret Subkey Packet", SecretSubkeyPacket),
     8:  ("Compressed Data Packet", None),
     9:  ("Symmetrically Encrypted Data Packet", None),
     10: ("Marker Packet", None),
@@ -518,6 +743,15 @@ TAG_TYPES = {
     63: ("Private", None),
 }
 
+S2K_TYPES = {
+    # (Name, Length)
+    0: ("Simple S2K", 2),
+    1: ("Salted S2K", 10),
+    2: ("Reserved value", 0),
+    3: ("Iterated and Salted S2K", 11),
+    101: ("GnuPG S2K", 6),
+}
+
 
 def new_tag_length(data, start):
     '''Takes a bytearray of data as input, as well as an offset of where to
